/*
 * dxr3audiodecoder.c
 *
 * Copyright (C) 2002-2004 Kai Möller
 * Copyright (C) 2004 Christian Gmeiner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


/*
  ToDo:
  - cDxr3AudioDecoder::Init: Why are we always reinit the codec?
*/

#include <stdio.h>
#include "dxr3audiodecoder.h"
#include "dxr3pesframe.h"

// ==================================
const int LPCM_HEADER_LENGTH = 7;

// ==================================
//! constructor
cDxr3AudioDecoder::cDxr3AudioDecoder() : rbuf(50000), ac3dtsDecoder(&rbuf)
{
    decoderOpened = false;
    volume = 255;
    Codec.id = CODEC_ID_MP2;

    Init();

    lastHeader[0] = 0xFF;
    lastHeader[1] = lastHeader[2] = lastHeader[3] = 0;
}

// ==================================
//! deconst.
cDxr3AudioDecoder::~cDxr3AudioDecoder()
{
    // close codec, if it is open
    cDxr3Ffmpeg::Instance().CloseCodec(Codec);
}

// ==================================
//! (re)init ffmpeg codec
void cDxr3AudioDecoder::Init()
{
    // (re)init codec
    cDxr3Ffmpeg::Instance().CloseCodec(Codec);
    if (cDxr3Ffmpeg::Instance().FindCodec(Codec))
    {
	cDxr3Ffmpeg::Instance().OpenCodec(Codec);
	rate = channels = -1;
	frameSize = Codec.codec_context->frame_size;
	decoderOpened = true;

	//lastHeader[0] = 0xFF;
	//lastHeader[1] = lastHeader[2] = lastHeader[3] = 0;
    }
    else
    {
	decoderOpened = false;
    }
}

// ==================================
//! decode given buffer
void cDxr3AudioDecoder::Decode(const uint8_t* buf, int length, uint32_t pts,
			       cDxr3SyncBuffer &aBuf)
{
    if (!decoderOpened)
    {
	// No decoder is open, so it
	// is better to stop here.
	return;
    }

    int len;
    int out_size;

    enum audioException
    {
	WRONG_LENGTH,
	UNEXPECTED_PARAMETER_CHANGE
    };

    // A PES packet doesn't necessarily starts/ends on an audio packet boundary,
    // so we have to buffer PES packets, decode the audio packets that are
    // complete and keep the rest.
    abuf.append((const char*)buf, length);
    buf = (uint8_t *) abuf.data();
    length = abuf.length();
    int i = 0;
    int fs = 0;
    for (i = 0; i < length-4 && !fs; i++)
    {
	if (checkMpegAudioHdr(buf+i))
	{
	    if ((buf[i+2] & 0xf0) != (lastHeader[2] & 0xf0))
	    {
		dsyslog("dxr3: audiodecoder: found different audio header"
			" (new: %#x, old: %#x), (re)initializing",
			*((uint32_t*) lastHeader), *((uint32_t*) (buf+i)));

		Init();
		lastHeader[0] = buf[i];
		lastHeader[1] = buf[i+1];
		lastHeader[2] = buf[i+2];
		lastHeader[3] = buf[i+3];
	    }
	    fs = calcFrameSize(buf+i);
	    if (i > 0)
	    {
	        abuf.erase(0, i);
	        buf = (uint8_t *) abuf.data();
	        length = abuf.length();
	    }
	}
    }

    if (!fs)
    {
	esyslog("dxr3: audiodecoder: no audio header found");
	abuf.clear();
	return;
    }

    try
    {
	int played = 0;
	while (length > 0 && length >= fs)
	{
	    out_size = AVCODEC_MAX_AUDIO_FRAME_SIZE;
#if LIBAVCODEC_VERSION_INT < ((51<<16)+(29<<8)+0)
	    len = avcodec_decode_audio(
#else
	    len = avcodec_decode_audio2(
#endif
		Codec.codec_context, (short *)(&pcmbuf), &out_size,
		const_cast<uint8_t *>(buf), fs);
	    if (len < 0 || out_size < 0)
		throw WRONG_LENGTH;

	    if (Codec.codec_context->sample_rate != rate)
	    {
		dsyslog("dxr3: audiodecoder: sample rate=%d",
			Codec.codec_context->sample_rate);
		if (rate != -1) throw UNEXPECTED_PARAMETER_CHANGE;
		rate = Codec.codec_context->sample_rate;
	    }
	    if (Codec.codec_context->channels != channels + 1)
	    {
		dsyslog("dxr3: audiodecoder: channels=%d",
			Codec.codec_context->channels);
		if (channels != -1)
		    throw UNEXPECTED_PARAMETER_CHANGE;
		channels = (Codec.codec_context->channels == 2) ? 1 : 0;
	    }
	    if (out_size)
	    {
		cFixedLengthFrame* pTempFrame = aBuf.Push(pcmbuf,
							  out_size, pts);
		if (pTempFrame)
		{
		    // TODO: should we break out of the loop on push timeout?
		    pTempFrame->SetChannelCount(channels);
		    pTempFrame->SetDataRate(rate);
		}
	    }
	    length -= len;
	    buf += len;
	    played += len;
	}
	if (played > 0)
	    abuf.erase(0, played);
    }
    catch (audioException ex)
    {
	switch (ex)
	{
	case WRONG_LENGTH:
	    esyslog("dxr3: audiodecoder: wrong length");
	    abuf.clear();
	    break;

	case UNEXPECTED_PARAMETER_CHANGE:
	    esyslog("dxr3: audiodecoder: unexpected parameter change");
	    abuf.clear();
	    break;

	default:
	    esyslog("dxr3: audiodecoder: unexpected exception");
	    abuf.clear();
	    break;
	}
	esyslog("dxr3: audiodecoder: skipping %d broken data bytes", length);

	Init();
    }
}

// ==================================
//! decode lpcm
void cDxr3AudioDecoder::DecodeLpcm(const uint8_t* buf, int length,
				   uint32_t pts, cDxr3SyncBuffer &aBuf)
{
    if (length > (LPCM_HEADER_LENGTH + 2))
    {
	// only even number of bytes are allowed
	if ((length - LPCM_HEADER_LENGTH) % 2 != 0)
	{
	    esyslog("dxr3: audiodecoder: skipping %d lpcm bytes", length);
	    return;
	}

	uint8_t* pFrame = new uint8_t[length - LPCM_HEADER_LENGTH];
	for (int i = LPCM_HEADER_LENGTH; i < length; i += 2)
	{
	    pFrame[i - LPCM_HEADER_LENGTH] = buf[i + 1];
	    pFrame[i - LPCM_HEADER_LENGTH + 1] = buf[i];
	}

	int codedSpeed = (buf[5] >> 4) & 0x03;
	int speed = 0;

	switch (codedSpeed)
	{
	case 1:
	    speed = 96000;
	    break;

	case 2:
	    speed = 44100;
	    break;

	case 3:
	    speed = 32000;
	    break;

	default:
	    speed = 48000;
	    break;
	}

	cFixedLengthFrame* pTempFrame = aBuf.Push(pFrame,
						  length - LPCM_HEADER_LENGTH,
						  pts);
	if (pTempFrame)
	{
	    pTempFrame->SetChannelCount(1);
	    pTempFrame->SetDataRate(speed);
	}
	delete[] pFrame;
    }
}

// ==================================
//! decode ac3
void cDxr3AudioDecoder::DecodeAc3Dts(const uint8_t* pPes, const uint8_t* buf,
				     int length, uint32_t pts,
				     cDxr3SyncBuffer &aBuf)
{
    int headerLength = (int) (buf - pPes);

    uint8_t* pBuf = (uint8_t*) pPes;
    ac3dtsDecoder.Check(pBuf + headerLength, length, pBuf);
    ac3dtsDecoder.Encapsulate(pBuf + headerLength, length);

    cFrame* pFrame = 0;
    while ((pFrame = rbuf.Get()))
    {
	if (pFrame && pFrame->Count())
	{
	    cDxr3PesFrame tempPes;
	    tempPes.ExtractNextFrame(pFrame->Data(), pFrame->Count());
	    int pesHeaderLength = (int) (tempPes.GetEsStart() - tempPes.GetPesStart());
	    uint8_t* pData = pFrame->Data() + pesHeaderLength + LPCM_HEADER_LENGTH;

	    for (int i = 0; i < pFrame->Count() - pesHeaderLength - LPCM_HEADER_LENGTH; i += 2)
	    {
		std::swap(pData[i], pData[i + 1]);
	    }

	    aBuf.Push(pFrame->Data() + pesHeaderLength + LPCM_HEADER_LENGTH, pFrame->Count() - pesHeaderLength - 7, tempPes.GetPts());
	    if (pFrame) rbuf.Drop(pFrame);
	}
    }
}

// ==================================
//! checking routine
bool cDxr3AudioDecoder::checkMpegAudioHdr(const uint8_t *head)
{
    // all informations about the mpeg audio header
    // can be found at http://www.datavoyage.com/mpgscript/mpeghdr.htm

    // the header conists of four 8bit elements, described as
    // AAAAAAAA AAABBCCD EEEEFFGH IIJJKLMM

    // all As must be set to 1
    if (head[0] != 0xff || (head[1] & 0xe0) != 0xe0)
    {
	return false;
    }

    // B is not allowed to be 01 (bits), as it is a reserved value
    if ((head[1] & 0x18) == 0x8)
    {
	return false;
    }

    // C is not allowed to be 00 (bits), as it is a reserved value
    if ((head[1] & 0x6) == 0x0)
    {
	return false;
    }

    // all Es are not allowed to be 1
    if ((head[2] & 0xf0) == 0xf0)
    {
	return false;
    }

    // all Fs are not allowed to be 1
    if ((head[2] & 0xc) == 0xc) {
        return false;
    }

    return true;
}

int cDxr3AudioDecoder::calcFrameSize(const uint8_t *header)
{
    static const int bitrates[2][3][15] =
    {
	{   // MPEG 1
	    {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,}, // Layer1
	    {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,}, // Layer2
	    {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,}  // Layer3
	},
	{   // MPEG 2, 2.5
	    {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},    // Layer1
	    {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},         // Layer2
	    {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,}          // Layer3
	}
    };

    static const int samplingRates[4][3] =
    {
	{11025, 12000, 8000,  },    // MPEG 2.5
	{0,     0,     0,     },    // reserved
	{22050, 24000, 16000, },    // MPEG 2
	{44100, 48000, 32000  }     // MPEG 1
    };

    static const int MPEG1 = 0x3;

    // in B the version is stored
    int ver = (header[1] >> 3) & 0x03;
    int version = 1;    // default to MPEG2

    // determine index into arrays based on ver
    if (ver == MPEG1)
    {
	version = 0;
    }

    // in C the layer version is stored
    int layer = 3 - ((header[1] >> 1) & 0x03);

    // in E the bitrate index is stored
    int bitrateIndex = (header[2] >> 4) & 0x0f;

    // in F the sampling rate frequency index is stored
    int samplingIndex = (header[2] >> 2) & 0x03;

    // in G the padding bit is stored
    int padding = (header[2] >> 1) & 0x01;

    int bitrate = bitrates[version][layer][bitrateIndex]; // kbit
    bitrate *= 1000; // kbit -> bit
    int samplesrate = samplingRates[ver][samplingIndex];

    // formulas used to calculate frame size
    //  layer I
    //      FrameLengthInBytes = (12 * BitRate / SampleRate + Padding) * 4
    //  layer II & III
    //      FrameLengthInBytes = 144 * BitRate / SampleRate + Padding
    if (layer == 0)
    {
	return (12 * bitrate / samplesrate + padding) * 4;
    }
    else
    {
	return 144 * bitrate / samplesrate + padding;
    }
}

// Local variables:
// mode: c++
// c-file-style: "stroustrup"
// c-file-offsets: ((inline-open . 0))
// c-basic-offset: 4
// indent-tabs-mode: t
// End:
